from __future__ import print_function
import numpy as np
from numpy import pi
from . import model

try:
    import mkl
    np.use_fastnumpy = True
except ImportError:
    pass

try:
    import pyfftw
    pyfftw.interfaces.cache.enable()
except ImportError:
    pass

class LayeredModel(model.Model):
    r"""Layered quasigeostrophic model.

    This model is meant to represent flows driven by baroclinic instabilty of a
    base-state shear. The potential vorticity anomalies qi are related to the
    streamfunction psii through

    .. math::

       {q_i} = \nabla^2\psi_i + \frac{f_0^2}{H_i} \left(\frac{\psi_{i-1}-
       \psi_i}{g'_{i-1}}- \frac{\psi_{i}-\psi_{i+1}}{g'_{i}}\right)\,,
       \qquad i = 2,\textsf{N}-1\,,

       {q_1} = \nabla^2\psi_1 + \frac{f_0^2}{H_1} \left(\frac{\psi_{2}-
       \psi_1}{g'_{1}}\right)\,,  \qquad i =1\,,

       {q_\textsf{N}} = \nabla^2\psi_\textsf{N} +
       \frac{f_0^2}{H_\textsf{N}} \left(\frac{\psi_{\textsf{N}-1}-
       \psi_\textsf{N}}{g'_{\textsf{N}}}\right) + \frac{f_0}{H_\textsf{N}}h_b\,,
       \qquad i =\textsf{N}\,,

    where the reduced gravity, or buoyancy jump, is

    .. math::

       g'_i \equiv g \frac{\rho_{i+1}-\rho_i}{\rho_i}\,.

    The evolution equations are

    .. math::

       \,{q_{i}}_t + \mathsf{J}\left(\psi_i\,, q_i\right) + \textsf{Q}_y {\psi_i}_x
       - \textsf{Q}_x {\psi_i}_y = \text{ssd} -
       r_{ek} \delta_{i\textsf{N}} \nabla^2 \psi_i\,, \qquad i = 1,\textsf{N}\,,


    where the mean potential vorticy gradients are

    .. math::

       \textsf{Q}_x = \textsf{S}\textsf{V}\,,

    and

    .. math::

       \textsf{Q}_y = \beta\,\textsf{I} - \textsf{S}\textsf{U}\,\,,

    where S is the stretching matrix, I is the identity matrix,
    and the background velocity is

    :math:`\vec{\textsf{V}}(z) = \left(\textsf{U},\textsf{V}\right)`.

    """

    def __init__(
        self,
        g = 9.81,
        beta=1.5e-11,               # gradient of coriolis parameter
        nz = 4,                     # number of layers
        rd=15000.0,                 # deformation radius
        H = None,                   # layer thickness
        U=None,                     # zonal base state flow
        V=None,                     # meridional base state flow
        rho = None,
        delta = None,
        **kwargs
        ):
        """
        Parameters
        ----------

        nz : integer number
             Number of layers (> 1)
        beta : number
            Gradient of coriolis parameter. Units: meters :sup:`-1`
            seconds :sup:`-1`
        rd : number
            Deformation radius. Units: meters. Only necessary for
            the two-layer (nz=2) case.
        delta : number
            Layer thickness ratio (H1/H2). Only necessary for the
            two-layer (nz=2) case. Unitless.
        U : list of size nz
            Base state zonal velocity. Units: meters s :sup:`-1`
        V : array of size nz
            Base state meridional velocity. Units: meters s :sup:`-1`
        H : array of size nz
            Layer thickness. Units: meters
        rho: array of size nz.
            Layer density. Units: kilograms meters :sup:`-3`

        """

        # physical
        self.g = g
        self.beta = beta
        self.rd = rd
        self.delta = delta
        self.nz = nz
        self.Ubg = np.array(U)
        self.Vbg = np.array(V)
        self.Hi = np.array(H)
        self.rhoi = np.array(rho)

        super(LayeredModel, self).__init__(**kwargs)

        self.vertical_modes()


    ### PRIVATE METHODS - not meant to be called by user ###


    def _initialize_stretching_matrix(self):
        """ Set up the stretching matrix """

        self.S = np.zeros((self.nz, self.nz))

        if (self.nz==2) and (self.rd) and (self.delta):

            self.del1 = self.delta/(self.delta+1.)
            self.del2 = (self.delta+1.)**-1
            self.Us = self.Ubg[0]-self.Ubg[1]

            self.F1 = self.rd**-2 / (1.+self.delta)
            self.F2 = self.delta*self.F1
            self.S[0,0], self.S[0,1] = -self.F1,  self.F1
            self.S[1,0], self.S[1,1] =  self.F2, -self.F2

        else:

            for i in range(self.nz):

                if i == 0:
                    self.S[i,i]   = -self.f2/self.Hi[i]/self.gpi[i]
                    self.S[i,i+1] =  self.f2/self.Hi[i]/self.gpi[i]

                elif i == self.nz-1:
                    self.S[i,i]   = -self.f2/self.Hi[i]/self.gpi[i-1]
                    self.S[i,i-1] =  self.f2/self.Hi[i]/self.gpi[i-1]

                else:
                    self.S[i,i-1] = self.f2/self.Hi[i]/self.gpi[i-1]
                    self.S[i,i]   = -(self.f2/self.Hi[i]/self.gpi[i] +
                                        self.f2/self.Hi[i]/self.gpi[i-1])
                    self.S[i,i+1] = self.f2/self.Hi[i]/self.gpi[i]

    def _initialize_background(self):
        """Set up background state (zonal flow and PV gradients)."""

        self.H = self.Hi.sum()

        if not (self.nz==2):
            self.gpi = self.g*(self.rhoi[1:]-self.rhoi[:-1])/self.rhoi[:-1]
            self.f2gpi = (self.f2/self.gpi)[:,np.newaxis,np.newaxis]

            assert self.gpi.size == self.nz-1, "Invalid size of gpi"

            assert np.all(self.gpi>0.), "Buoyancy jump has negative sign!"

            assert self.Hi.size == self.nz, self.logger.error('size of Hi does not' +
                     'match number of vertical levels nz')

            assert self.rhoi.size == self.nz, self.logger.error('size of rhoi does not' +
                     'match number of vertical levels nz')

            assert self.Ubg.size == self.nz, self.logger.error('size of Ubg does not' +
                     'match number of vertical levels nz')

            assert self.Vbg.size == self.nz, self.logger.error('size of Vbg does not' +
                     'match number of vertical levels nz')

        else:
            self.f2gpi = np.array(self.rd**-2 *
                (self.Hi[0]*self.Hi[1])/self.H)[np.newaxis,np.newaxis]


        self._initialize_stretching_matrix()

        # the meridional PV gradients in each layer
        self.Qy = self.beta - np.dot(self.S,self.Ubg)
        self.Qx = np.dot(self.S,self.Vbg)


        # complex versions, multiplied by k, speeds up computations to precompute
        self.ikQy = self.Qy[:,np.newaxis,np.newaxis]*1j*self.k
        self.ilQx = self.Qx[:,np.newaxis,np.newaxis]*1j*self.l

    def _initialize_inversion_matrix(self):

        a = np.ma.zeros((self.nz, self.nz, self.nl, self.nk), np.dtype('float64'))

        if (self.nz==2):
            det_inv =  np.ma.masked_equal(
                    ( (self.S[0,0]-self.wv2)*(self.S[1,1]-self.wv2) -\
                            self.S[0,1]*self.S[1,0] ), 0.)**-1
            a[0,0] = (self.S[1,1]-self.wv2)*det_inv
            a[0,1] = -self.S[0,1]*det_inv
            a[1,0] = -self.S[1,0]*det_inv
            a[1,1] = (self.S[0,0]-self.wv2)*det_inv
        else:
            I = np.eye(self.nz)[:,:,np.newaxis,np.newaxis]
            M = self.S[:,:,np.newaxis,np.newaxis]-I*self.wv2
            M[:,:,0,0] = np.nan  # avoids singular matrix in inv()
            a = np.linalg.inv(M.T).T

        self.a = np.ma.masked_invalid(a).filled(0.)

    def _initialize_forcing(self):
        pass
        #"""Set up frictional filter."""
        # this defines the spectral filter (following Arbic and Flierl, 2003)
        # cphi=0.65*pi
        # wvx=np.sqrt((self.k*self.dx)**2.+(self.l*self.dy)**2.)
        # self.filtr = np.exp(-self.filterfac*(wvx-cphi)**4.)
        # self.filtr[wvx<=cphi] = 1.

    ### All the diagnostic stuff follows. ###
    def _calc_cfl(self):
        return np.abs(
            np.hstack([self.u + self.Ubg[:,np.newaxis,np.newaxis], self.v])
        ).max()*self.dt/self.dx

    # calculate KE: this has units of m^2 s^{-2}
    #   (should also multiply by H1 and H2...)
    def _calc_ke(self):
        ke = 0.
        for j in range(self.nz):
            ke += .5*self.Hi[j]*self.spec_var(self.wv*self.ph[j])
        return ke.sum() / self.H

    # calculate eddy turn over time
    # (perhaps should change to fraction of year...)
    def _calc_eddy_time(self):
        """ estimate the eddy turn-over time in days """
        ens = 0.
        for j in range(self.nz):
            ens = .5*self.Hi[j] * self.spec_var(self.wv2*self.ph[j])

        return 2.*pi*np.sqrt( self.H / ens.sum() ) / 86400

    def _calc_derived_fields(self):

        self.p = self.ifft(self.ph)
        self.xi =self.ifft(-self.wv2*self.ph)
        self.Jpxi = self._advect(self.xi, self.u, self.v)
        self.Jq = self._advect(self.q, self.u, self.v)

        self.Sph = np.einsum("ij,jkl->ikl",self.S,self.ph)
        self.Sp = self.ifft(self.Sph)
        self.JSp = self._advect(self.Sp,self.u,self.v)

        self.phn = self.modal_projection(self.ph)


    def _initialize_model_diagnostics(self):
        """ Extra diagnostics for layered model """

        self.add_diagnostic('entspec',
                description='barotropic enstrophy spectrum',
                function= (lambda self:
                    np.abs((self.Hi[:,np.newaxis,np.newaxis]*self.qh).sum(axis=0))**2/self.H) )

        self.add_diagnostic('KEspec_modal',
                description='modal KE spectra',
                function= (lambda self:
                    self.wv2*(np.abs(self.phn)**2)/self.M**2 ))

        self.add_diagnostic('PEspec_modal',
                description='modal PE spectra',
                function= (lambda self:
                    self.kdi2[1:,np.newaxis,np.newaxis]*(np.abs(self.phn[1:,:,:])**2)/self.M**2 ))

        self.add_diagnostic('APEspec',
                description='available potential energy spectrum',
                function= (lambda self:
                           (self.f2gpi*
                            np.abs(self.ph[:-1]-self.ph[1:])**2).sum(axis=0)/self.H))

        self.add_diagnostic('KEflux',
                    description='spectral divergence of flux of kinetic energy',
                    function =(lambda self: (self.Hi[:,np.newaxis,np.newaxis]*
                               (self.ph.conj()*self.Jpxi).real).sum(axis=0)/self.H))

        self.add_diagnostic('APEflux',
                    description='spectral divergence of flux of available potential energy',
                    function =(lambda self: (self.Hi[:,np.newaxis,np.newaxis]*
                               (self.ph.conj()*self.JSp).real).sum(axis=0)/self.H))

        self.add_diagnostic('APEgenspec',
                    description='the spectrum of the rate of generation of available potential energy',
                    function =(lambda self: (self.Hi[:,np.newaxis,np.newaxis]*
                                (self.Ubg[:,np.newaxis,np.newaxis]*self.k +
                                 self.Vbg[:,np.newaxis,np.newaxis]*self.l)*
                                (1j*self.ph.conj()*self.Sph).real).sum(axis=0)/self.H))

        self.add_diagnostic('ENSflux',
                 description='barotropic enstrophy flux',
                 function = (lambda self: (-self.Hi[:,np.newaxis,np.newaxis]*
                              (self.qh.conj()*self.Jq).real).sum(axis=0)/self.H))

        self.add_diagnostic('ENSgenspec',
                    description='the spectrum of the rate of generation of barotropic enstrophy',
                    function = (lambda self:
                            -(self.Hi[:,np.newaxis,np.newaxis]*((self.ikQy -
                            self.ilQx)*(self.Sph.conj()*self.ph)).real).sum(axis=0)/self.H))



from __future__ import print_function
import numpy as np
from numpy import pi
from . import model


class SQGModel(model.Model):
    """Surface quasigeostrophic model."""

    def __init__(
        self,
        beta=0.,                    # gradient of coriolis parameter
        Nb = 1.,                    # Buoyancy frequency
        rd=0.,                      # deformation radius
        H = 1.,                     # depth of layer
        U=0.,                       # max vel. of base-state
        **kwargs
        ):
        """
        Parameters
        ----------

        beta : number
            Gradient of coriolis parameter. Units: meters :sup:`-1`
            seconds :sup:`-1`
        Nb : number
            Buoyancy frequency. Units: seconds :sup:`-1`.
        U : number
            Background zonal flow. Units: meters.
        """

        # physical
        self.beta = beta
        self.Nb = Nb
        #self.rek = rek
        self.rd = rd
        self.H = H
        self.Hi = np.array(H)[np.newaxis,...]
        self.U = U
        #self.filterfac = filterfac

        self.nz = 1

        super(SQGModel, self).__init__(**kwargs)

        # initial conditions: (PV anomalies)
        self.set_q(1e-3*np.random.rand(1,self.ny,self.nx))

    ### PRIVATE METHODS - not meant to be called by user ###

    def _initialize_background(self):
        """Set up background state (zonal flow and PV gradients)."""

        # the meridional PV gradients in each layer
        self.Qy = np.asarray(self.beta)[np.newaxis, ...]

        # background vel.
        self.set_U(self.U)

        # complex versions, multiplied by k, speeds up computations to pre-compute
        self.ikQy = self.Qy * 1j * self.k

        self.ilQx = 0.

    def _initialize_inversion_matrix(self):
        """ the inversion """
        # The sqg model is diagonal. The inversion is simply qh = -kappa**2 ph
        self.a = np.asarray(self.Nb*np.sqrt(self.wv2i))[np.newaxis, np.newaxis, :, :]

    def _initialize_forcing(self):
        pass

    def set_U(self, U):
        """Set background zonal flow"""
        self.Ubg = np.asarray(U)[np.newaxis,...]

    def _calc_diagnostics(self):
        # here is where we calculate diagnostics
        if (self.t>=self.dt) and (self.tc%self.taveints==0):
            self._increment_diagnostics()

    ### All the diagnostic stuff follows. ###
    def _calc_cfl(self):
        return np.abs(
            np.hstack([self.u + self.Ubg, self.v])
        ).max()*self.dt/self.dx

    # calculate KE: this has units of m^2 s^{-2}
    def _calc_ke(self):
        ke = .5*self.spec_var(self.wv*self.ph)
        return ke.sum()

    # calculate eddy turn over time
    # (perhaps should change to fraction of year...)
    def _calc_eddy_time(self):
        """ estimate the eddy turn-over time in days """
        ens = .5*self.H * spec_var(self, self.wv2*self.ph)
        return 2.*pi*np.sqrt( self.H / ens ) / year



from __future__ import print_function
import numpy as np
from numpy import pi
from . import model

try:
    import mkl
    np.use_fastnumpy = True
except ImportError:
    pass

try:
    import pyfftw
    pyfftw.interfaces.cache.enable()
except ImportError:
    pass

class QGModel(model.Model):
    r"""Two layer quasigeostrophic model.

    This model is meant to representflows driven by baroclinic instabilty of a
    base-state shear :math:`U_1-U_2`. The upper and lower
    layer potential vorticity anomalies :math:`q_1` and :math:`q_2` are

    .. math::

        q_1 &= \nabla^2\psi_1 + F_1(\psi_2 - \psi_1) \\
        q_2 &= \nabla^2\psi_2 + F_2(\psi_1 - \psi_2)

    with

    .. math::
        F_1 &\equiv \frac{k_d^2}{1 + \delta^2} \\
        F_2 &\equiv \delta F_1 \ .

    The layer depth ratio is given by :math:`\delta = H_1 / H_2`.
    The total depth is :math:`H = H_1 + H_2`.

    The background potential vorticity gradients are

    .. math::

        \beta_1 &= \beta + F_1(U_1 - U_2) \\
        \beta_2 &= \beta - F_2( U_1 - U_2) \ .

    The evolution equations for :math:`q_1` and :math:`q_2` are

    .. math::

        \partial_t\,{q_1} + J(\psi_1\,, q_1) + \beta_1\,
        {\psi_1}_x &= \text{ssd} \\
        \partial_t\,{q_2} + J(\psi_2\,, q_2)+ \beta_2\, {\psi_2}_x
        &= -r_{ek}\nabla^2 \psi_2 + \text{ssd}\,.

    where `ssd` represents small-scale dissipation and :math:`r_{ek}` is the
    Ekman friction parameter.

    """


    def __init__(
        self,
        beta=1.5e-11,               # gradient of coriolis parameter
        #rek=5.787e-7,               # linear drag in lower layer
        rd=15000.0,                 # deformation radius
        delta=0.25,                 # layer thickness ratio (H1/H2)
        H1 = 500,                   # depth of layer 1 (H1)
        U1=0.025,                   # upper layer flow
        U2=0.0,                     # lower layer flow
        **kwargs
        ):
        """
        Parameters
        ----------

        beta : number
            Gradient of coriolis parameter. Units: meters :sup:`-1`
            seconds :sup:`-1`
        rek : number
            Linear drag in lower layer. Units: seconds :sup:`-1`
        rd : number
            Deformation radius. Units: meters.
        delta : number
            Layer thickness ratio (H1/H2)
        U1 : number
            Upper layer flow. Units: m/s
        U2 : number
            Lower layer flow. Units: m/s
        """

        # physical
        self.beta = beta
        #self.rek = rek
        self.rd = rd
        self.delta = delta
        self.Hi = np.array([ H1, H1/delta])
        self.U1 = U1
        self.U2 = U2
        #self.filterfac = filterfac

        self.nz = 2

        super(QGModel, self).__init__(**kwargs)

        # initial conditions: (PV anomalies)
        self.set_q1q2(
            1e-7*np.random.rand(self.ny,self.nx) + 1e-6*(
                np.ones((self.ny,1)) * np.random.rand(1,self.nx) ),
                np.zeros_like(self.x) )



    ### PRIVATE METHODS - not meant to be called by user ###

    def _initialize_background(self):
        """Set up background state (zonal flow and PV gradients)."""

        # Background zonal flow (m/s):
        self.H = self.Hi.sum()
        self.set_U1U2(self.U1, self.U2)
        self.U = self.U1 - self.U2

        # the F parameters
        self.F1 = self.rd**-2 / (1.+self.delta)
        self.F2 = self.delta*self.F1

        # the meridional PV gradients in each layer
        self.Qy1 = self.beta + self.F1*(self.U1 - self.U2)
        self.Qy2 = self.beta - self.F2*(self.U1 - self.U2)
        self.Qy = np.array([self.Qy1, self.Qy2])
        # complex versions, multiplied by k, speeds up computations to precompute
        self.ikQy1 = self.Qy1 * 1j * self.k
        self.ikQy2 = self.Qy2 * 1j * self.k

        # vector version
        self.ikQy = np.vstack([self.ikQy1[np.newaxis,...],
                               self.ikQy2[np.newaxis,...]])
        self.ilQx = 0.

        # layer spacing
        self.del1 = self.delta/(self.delta+1.)
        self.del2 = (self.delta+1.)**-1

    def _initialize_inversion_matrix(self):

        # The matrix multiplication will look like this
        # ph[0] = a[0,0] * self.qh[0] + a[0,1] * self.qh[1]
        # ph[1] = a[1,0] * self.qh[0] + a[1,1] * self.qh[1]

        a = np.ma.zeros((self.nz, self.nz, self.nl, self.nk), np.dtype('float64'))
        # inverse determinant
        det_inv =  np.ma.masked_equal(
                self.wv2 * (self.wv2 + self.F1 + self.F2), 0.)**-1
        a[0,0] = -(self.wv2 + self.F2)*det_inv
        a[0,1] = -self.F1*det_inv
        a[1,0] = -self.F2*det_inv
        a[1,1] = -(self.wv2 + self.F1)*det_inv

        self.a = np.ma.masked_invalid(a).filled(0.)

    def _initialize_forcing(self):
        pass
        #"""Set up frictional filter."""
        # this defines the spectral filter (following Arbic and Flierl, 2003)
        # cphi=0.65*pi
        # wvx=np.sqrt((self.k*self.dx)**2.+(self.l*self.dy)**2.)
        # self.filtr = np.exp(-self.filterfac*(wvx-cphi)**4.)
        # self.filtr[wvx<=cphi] = 1.

    def set_q1q2(self, q1, q2, check=False):
        """Set upper and lower layer PV anomalies.

        Parameters
        ----------

        q1 : array-like
            Upper layer PV anomaly in spatial coordinates.
        q1 : array-like
            Lower layer PV anomaly in spatial coordinates.
        """
        self.set_q(np.vstack([q1[np.newaxis,:,:], q2[np.newaxis,:,:]]))
        #self.q[0] = q1
        #self.q[1] = q2

        # initialize spectral PV
        #self.qh = self.fft2(self.q)

        # check that it works
        if check:
            np.testing.assert_allclose(self.q1, q1)
            np.testing.assert_allclose(self.q1, self.ifft2(self.qh1))

    def set_U1U2(self, U1, U2):
        """Set background zonal flow.

        Parameters
        ----------

        U1 : number
            Upper layer flow. Units: m/s
        U2 : number
            Lower layer flow. Units: m/s
        """
        self.U1 = U1
        self.U2 = U2
        #self.Ubg = np.array([U1,U2])[:,np.newaxis,np.newaxis]
        self.Ubg = np.array([U1,U2])

    ### All the diagnostic stuff follows. ###
    def _calc_cfl(self):
        return np.abs(
            np.hstack([self.u + self.Ubg[:,np.newaxis,np.newaxis], self.v])
        ).max()*self.dt/self.dx

    # calculate KE: this has units of m^2 s^{-2}
    #   (should also multiply by H1 and H2...)
    def _calc_ke(self):
        ke1 = .5*self.Hi[0]*self.spec_var(self.wv*self.ph[0])
        ke2 = .5*self.Hi[1]*self.spec_var(self.wv*self.ph[1])
        return ( ke1.sum() + ke2.sum() ) / self.H

    # calculate eddy turn over time
    # (perhaps should change to fraction of year...)
    def _calc_eddy_time(self):
        """ estimate the eddy turn-over time in days """

        ens = .5*self.Hi[0] * self.spec_var(self.wv2*self.ph1) + \
            .5*self.Hi[1] * self.spec_var(self.wv2*self.ph2)

        return 2.*pi*np.sqrt( self.H / ens ) / 86400

    def _calc_derived_fields(self):
        self.p = self.ifft(self.ph)
        self.xi =self.ifft(-self.wv2*self.ph)
        self.Jptpc = -self._advect(
                    (self.p[0] - self.p[1]),
                    (self.del1*self.u[0] + self.del2*self.u[1]),
                    (self.del1*self.v[0] + self.del2*self.v[1]))
        # fix for delta.neq.1
        self.Jpxi = self._advect(self.xi, self.u, self.v)

    def _initialize_model_diagnostics(self):
        """Extra diagnostics for two-layer model"""

        self.add_diagnostic('entspec',
            description='barotropic enstrophy spectrum',
            function= (lambda self:
                      np.abs(self.del1*self.qh[0] + self.del2*self.qh[1])**2.)
        )

        self.add_diagnostic('APEflux',
            description='spectral flux of available potential energy',
            function= (lambda self:
              self.rd**-2 * self.del1*self.del2 *
              np.real((self.ph[0]-self.ph[1])*np.conj(self.Jptpc)) )
        )

        self.add_diagnostic('KEflux',
            description='spectral flux of kinetic energy',
            function= (lambda self:
              np.real(self.del1*self.ph[0]*np.conj(self.Jpxi[0])) +
              np.real(self.del2*self.ph[1]*np.conj(self.Jpxi[1])) )
        )

        self.add_diagnostic('APEgenspec',
            description='spectrum of APE generation',
            function= (lambda self: self.U * self.rd**-2 * self.del1 * self.del2 *
                       np.real(1j*self.k*(self.del1*self.ph[0] + self.del2*self.ph[0]) *
                                  np.conj(self.ph[0] - self.ph[1])) )
        )

        self.add_diagnostic('APEgen',
            description='total APE generation',
            function= (lambda self: self.U * self.rd**-2 * self.del1 * self.del2 *
                       np.real((1j*self.k*
                            (self.del1*self.ph[0] + self.del2*self.ph[1]) *
                            np.conj(self.ph[0] - self.ph[1])).sum()
                              +(1j*self.k[:,1:-2]*
                            (self.del1*self.ph[0,:,1:-2] + self.del2*self.ph[1,:,1:-2]) *
                            np.conj(self.ph[0,:,1:-2] - self.ph[1,:,1:-2])).sum()) /
                            (self.M**2) )
        )

        ### These generic diagnostics are now calculated in model.py ###
        # self.add_diagnostic('KE1spec',
        #     description='upper layer kinetic energy spectrum',
        #     function=(lambda self: 0.5*self.wv2*np.abs(self.ph[0])**2)
        # )
        #
        # self.add_diagnostic('KE2spec',
        #     description='lower layer kinetic energy spectrum',
        #     function=(lambda self: 0.5*self.wv2*np.abs(self.ph[1])**2)
        # )
        #
        # self.add_diagnostic('q1',
        #     description='upper layer QGPV',
        #     function= (lambda self: self.q[0])
        # )
        #
        # self.add_diagnostic('q2',
        #     description='lower layer QGPV',
        #     function= (lambda self: self.q[1])
        # )
        #
        # self.add_diagnostic('EKE1',
        #     description='mean upper layer eddy kinetic energy',
        #     function= (lambda self: 0.5*(self.v[0]**2 + self.u[0]**2).mean())
        # )
        #
        # self.add_diagnostic('EKE2',
        #     description='mean lower layer eddy kinetic energy',
        #     function= (lambda self: 0.5*(self.v[1]**2 + self.u[1]**2).mean())
        # )
        #
        # self.add_diagnostic('EKEdiss',
        #     description='total energy dissipation by bottom drag',
        #     function= (lambda self:
        #                (self.del2*self.rek*self.wv2*
        #                 np.abs(self.ph[1])**2./(self.nx*self.ny)).sum())
        # )



from __future__ import print_function
"""Utility functions for pyqg model data."""

import numpy as np
from numpy import pi

def spec_var(model, ph):
    """Compute variance of ``p`` from Fourier coefficients ``ph``.

    Parameters
    ----------
    model : pyqg.Model instance
        The model object from which `ph` originates
    ph : complex array
        The field on which to compute the variance

    Returns
    -------
    var_dens : float
        The variance of `ph`
    """

    var_dens = 2. * np.abs(ph)**2 / model.M**2
    # only half of coefs [0] and [nx/2+1] due to symmetry in real fft2
    var_dens[...,0] /= 2
    var_dens[...,-1] /= 2
    return var_dens.sum(axis=(-1,-2))


def spec_sum(ph2):
    """Compute total spectral sum of the real spectral quantity``ph^2``.

    Parameters
    ----------
    model : pyqg.Model instance
        The model object from which `ph` originates
    ph2 : real array
        The field on which to compute the sum

    Returns
    -------
    var_dens : float
        The sum of `ph2`
    """

    ph2 = 2.*ph2
    ph2[...,0] = ph2[...,0]/2.
    ph2[...,-1] = ph2[...,-1]/2.

    return ph2.sum(axis=(-1,-2))


def calc_ispec(model, ph):
    """Compute isotropic spectrum `phr` of `ph` from 2D spectrum.

    Parameters
    ----------
    model : pyqg.Model instance
        The model object from which `ph` originates
    ph : complex array
        The field on which to compute the variance

    Returns
    -------
    kr : array
        isotropic wavenumber
    phr : array
        isotropic spectrum
    """

    if model.kk.max()>model.ll.max():
        kmax = model.ll.max()
    else:
        kmax = model.kk.max()

    # create radial wavenumber
    dkr = np.sqrt(model.dk**2 + model.dl**2)
    kr =  np.arange(dkr/2.,kmax+dkr,dkr)
    phr = np.zeros(kr.size)

    for i in range(kr.size):
        fkr =  (model.wv>=kr[i]-dkr/2) & (model.wv<=kr[i]+dkr/2)
        dth = pi / (fkr.sum()-1)
        phr[i] = ph[fkr].sum() * kr[i] * dth

    return kr, phr



#cython: profile=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False 
#from __future__ import division
import numpy as np
import warnings
import cython
cimport numpy as np
from cython.parallel import prange, threadid

# see if we got a compile time flag
include '.compile_time_use_pyfftw.pxi'
IF PYQG_USE_PYFFTW:
    import pyfftw
    pyfftw.interfaces.cache.enable()  
ELSE:
    import numpy.fft as npfft
    warnings.warn('No pyfftw detected. Using numpy.fft') 

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
DTYPE_real = np.float64
DTYPE_com = np.complex128
# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.float64_t DTYPE_real_t
ctypedef np.complex128_t DTYPE_com_t

cdef class PseudoSpectralKernel:
    # array shapes
    cdef public int Nx, Ny, Nz
    cdef public int Nk, Nl
    
    ### the main state variables (memory views to numpy arrays) ###
    # pv
    cdef DTYPE_real_t [:, :, :] q
    cdef DTYPE_com_t [:, :, :] qh
    # streamfunction
    cdef DTYPE_com_t [:, :, :] ph
    # velocities
    cdef DTYPE_real_t [:, :, :] u
    cdef DTYPE_real_t [:, :, :] v
    cdef DTYPE_com_t [:, :, :] uh
    cdef DTYPE_com_t [:, :, :] vh
    # pv fluxes
    cdef DTYPE_real_t [:, :, :] uq
    cdef DTYPE_real_t [:, :, :] vq
    cdef DTYPE_com_t [:, :, :] uqh
    cdef DTYPE_com_t [:, :, :] vqh
    # the tendencies
    cdef DTYPE_com_t [:, :, :] dqhdt
    cdef DTYPE_com_t [:, :, :] dqhdt_p
    cdef DTYPE_com_t [:, :, :] dqhdt_pp
    
    # dummy variables for diagnostic ffts
    cdef DTYPE_real_t [:, :, :] _dummy_fft_in
    cdef DTYPE_com_t [:, :, :] _dummy_fft_out
    cdef DTYPE_real_t [:, :, :] _dummy_ifft_out
    cdef DTYPE_com_t [:, :, :] _dummy_ifft_in
    
    # the variables needed for inversion and advection
    # store a as complex so we don't have to typecast in inversion
    cdef DTYPE_com_t [:, :, :, :] a
    cdef DTYPE_com_t [:] _ik
    cdef DTYPE_com_t [:] _il
    cdef public DTYPE_real_t [:,:] _k2l2
    # background state constants (functions of z only)
    cdef DTYPE_real_t [:] Ubg
    cdef DTYPE_com_t [:, :] _ikQy
    # spectral filter
    cdef public DTYPE_real_t [:, :] _filtr
    
    # friction parameter
    cdef public DTYPE_real_t _rek
    
    # time
    cdef public int tc
    cdef public DTYPE_real_t _dt
    cdef public DTYPE_real_t t
    
    # threading
    cdef int num_threads
    # number of elements per work group in the y / l direction
    cdef int chunksize

    # pyfftw objects (callable)
    cdef object fft_q_to_qh
    cdef object ifft_qh_to_q
    cdef object ifft_uh_to_u
    cdef object ifft_vh_to_v
    cdef object fft_uq_to_uqh
    cdef object fft_vq_to_vqh
    cdef object _dummy_fft
    cdef object _dummy_ifft
        
    def _kernel_init(self, int Nz, int Ny, int Nx, 
                    np.ndarray[DTYPE_real_t, ndim=4] a,
                    np.ndarray[DTYPE_real_t, ndim=1] k,
                    np.ndarray[DTYPE_real_t, ndim=1] l,
                    np.ndarray[DTYPE_real_t, ndim=1] Ubg,
                    np.ndarray[DTYPE_real_t, ndim=1] Qy,
                    np.ndarray[DTYPE_real_t, ndim=2] filtr,
                    DTYPE_real_t dt=1.0,
                    DTYPE_real_t rek=0.0,
                    fftw_num_threads=1,                                       
    ):
        self.Nz = Nz
        self.Ny = Ny
        self.Nx = Nx
        self.Nl = Ny
        self.Nk = Nx/2 + 1
        
        self._rek = rek
        
        ### none of this shape checking works
        #assert a.shape == (self.Nz, self.Nz self.Nl, self.Nk):
        #assert k.shape == (self.Nk,):
        #assert l.shape == (self.Nl,):
        #assert Ubg.shape == (self.Nz,), 'Ubg is the wrong shape'
        #assert Vbg.shape == (self.Nz,), 'Vbg is the wrong shape'
        #assert Qx.shape == (self.Nz,), 'Qx is the wrong shape'
        #assert Qy.shape == (self.Nz,), 'Qy is the wrong shape'
        
        # assign a, _ik, _il
        self.a = a.astype(DTYPE_com)
        self._ik = 1j*k
        self._il = 1j*l
        
        self._k2l2 = np.zeros((self.Nl, self.Nk), DTYPE_real)
        for j in range(self.Nl):
            for i in range(self.Nk):
                self._k2l2[j,i] = k[i]**2 + l[j]**2
        
        # assign Ubg, Vbg, _ilQx, _ikQy
        self.Ubg = Ubg
        #self.Vbg = Vbg
        self._ikQy = 1j * k[np.newaxis, :] * Qy[:, np.newaxis]
        
        # initialize FFT inputs / outputs as byte aligned by pyfftw
        q = self._empty_real()
        self.q = q # assign to memory view
        qh = self._empty_com()
        self.qh = qh
        
        ph = self._empty_com()
        self.ph = ph
        
        u = self._empty_real()
        self.u = u
        uh = self._empty_com()
        self.uh = uh
        
        v = self._empty_real()
        self.v = v
        vh = self._empty_com()
        self.vh = vh
        
        uq = self._empty_real()
        self.uq = uq
        uqh = self._empty_com()
        self.uqh = uqh
        
        vq = self._empty_real()
        self.vq = vq
        vqh = self._empty_com()
        self.vqh = vqh
        
        # dummy variables for diagnostic ffts
        dfftin = self._empty_real()
        self._dummy_fft_in = dfftin
        dfftout = self._empty_com()
        self._dummy_fft_out = dfftout
        difftin = self._empty_com()
        self._dummy_ifft_in = difftin
        difftout = self._empty_real()
        self._dummy_ifft_out = difftout
        
        # the tendency
        self.dqhdt = self._empty_com()
        self.dqhdt_p = self._empty_com()
        self.dqhdt_pp = self._empty_com()
        
        # spectral filter
        self._filtr = filtr
        
        # time stuff
        self._dt = dt
        self.tc = 0
        self.t = 0.0
        
        # for threading
        self.num_threads = fftw_num_threads
        self.chunksize = self.Nl/self.num_threads
        
        IF PYQG_USE_PYFFTW:
            # set up FFT plans
            # Note that the Backwards Real transform for the case
            # in which the dimensionality of the transform is greater than 1
            # will destroy the input array. This is inherent to FFTW and the only
            # general work-around for this is to copy the array prior to
            # performing the transform.
            self.fft_q_to_qh = pyfftw.FFTW(q, qh, threads=fftw_num_threads,
                             direction='FFTW_FORWARD', axes=(-2,-1))
            self.ifft_qh_to_q = pyfftw.FFTW(qh, q, threads=fftw_num_threads,
                             direction='FFTW_BACKWARD', axes=(-2,-1))
            self.ifft_uh_to_u = pyfftw.FFTW(uh, u, threads=fftw_num_threads, 
                             direction='FFTW_BACKWARD', axes=(-2,-1))
            self.ifft_vh_to_v = pyfftw.FFTW(vh, v, threads=fftw_num_threads, 
                             direction='FFTW_BACKWARD', axes=(-2,-1))
            self.fft_uq_to_uqh = pyfftw.FFTW(uq, uqh, threads=fftw_num_threads, 
                             direction='FFTW_FORWARD', axes=(-2,-1))
            self.fft_vq_to_vqh = pyfftw.FFTW(vq, vqh, threads=fftw_num_threads, 
                             direction='FFTW_FORWARD', axes=(-2,-1))
            # dummy ffts for diagnostics
            self._dummy_fft = pyfftw.FFTW(dfftin, dfftout, threads=fftw_num_threads, 
                             direction='FFTW_FORWARD', axes=(-2,-1))
            self._dummy_ifft = pyfftw.FFTW(difftin, difftout, threads=fftw_num_threads, 
                             direction='FFTW_BACKWARD', axes=(-2,-1))
    
    # otherwise define those functions using numpy
    IF PYQG_USE_PYFFTW==0:
        def fft_q_to_qh(self):
            self.qh = npfft.rfftn(self.q, axes=(-2,-1))
        def ifft_qh_to_q(self):
            self.q = npfft.irfftn(self.qh, axes=(-2,-1))
        def ifft_uh_to_u(self):
            self.u = npfft.irfftn(self.uh, axes=(-2,-1))
        def ifft_vh_to_v(self):
            self.v = npfft.irfftn(self.vh, axes=(-2,-1))
        def fft_uq_to_uqh(self):
            self.uqh = npfft.rfftn(self.uq, axes=(-2,-1))
        def fft_vq_to_vqh(self):
            self.vqh = npfft.rfftn(self.vq, axes=(-2,-1))
        def _dummy_fft(self):
            self._dummy_fft_out = npfft.rfftn(self._dummy_fft_in, axes=(-2,-1))
        def _dummy_ifft(self):
            self._dummy_ifft_out = npfft.irfftn(self._dummy_ifft_in, axes=(-2,-1))
            
    def _empty_real(self):
        """Allocate a space-grid-sized variable for use with fftw transformations."""
        shape = (self.Nz, self.Ny, self.Ny)
        IF PYQG_USE_PYFFTW:
            return pyfftw.n_byte_align_empty(shape,
                                 pyfftw.simd_alignment, dtype=DTYPE_real)
        ELSE:
            return np.empty(shape, dtype=DTYPE_real)            

    def _empty_com(self):
        """Allocate a Fourier-grid-sized variable for use with fftw transformations."""
        shape = (self.Nz, self.Nl, self.Nk)
        IF PYQG_USE_PYFFTW:
            return pyfftw.n_byte_align_empty(shape,
                                 pyfftw.simd_alignment, dtype=DTYPE_com)
        ELSE:
            return np.empty(shape, dtype=DTYPE_com)
    
    def fft(self, np.ndarray[DTYPE_real_t, ndim=3] v):
        """"Generic FFT function for real grid-sized variables.
        Not used for actual model ffs."""
        cdef  DTYPE_real_t [:, :, :] v_view = v
        # copy input into memory view
        self._dummy_fft_in[:] = v_view
        self._dummy_fft()
        # return a copy of the output
        return np.asarray(self._dummy_fft_out).copy()

    def ifft(self, np.ndarray[DTYPE_com_t, ndim=3] v):
        """"Generic IFFT function for complex grid-sized variables.
        Not used for actual model ffs."""
        cdef  DTYPE_com_t [:, :, :] v_view = v
        # copy input into memory view
        self._dummy_ifft_in[:] = v_view
        self._dummy_ifft()
        # return a copy of the output
        return np.asarray(self._dummy_ifft_out).copy()
   
    # the only way to set q and qh
    def set_qh(self, np.ndarray[DTYPE_com_t, ndim=3] b):
        cdef  DTYPE_com_t [:, :, :] b_view = b
        self.qh[:] = b_view
        self.ifft_qh_to_q()
        # input might have been destroyed, have to re-assign
        self.qh[:] = b_view
        
    def set_q(self, np.ndarray[DTYPE_real_t, ndim=3] b):
        cdef  DTYPE_real_t [:, :, :] b_view = b
        self.q[:] = b_view
        self.fft_q_to_qh()

    def _invert(self):
        self.__invert()

    cdef void __invert(self) nogil:
        ### algorithm
        # invert ph = a * qh
        # uh, vh = -_il * ph, _ik * ph
        # u, v, = ifft(uh), ifft(vh)

        cdef Py_ssize_t k, k1, k2, j, i
        # set ph to zero
        for k in range(self.Nz):
            for j in prange(self.Nl, nogil=True, schedule='static',
                      chunksize=self.chunksize,  
                      num_threads=self.num_threads):
                for i in range(self.Nk):
                    self.ph[k,j,i] = (0. + 0.*1j) 
                    
        # invert qh to find ph
        for k2 in range(self.Nz):
            for k1 in range(self.Nz):
                for j in prange(self.Nl, nogil=True, schedule='static',
                          chunksize=self.chunksize,  
                          num_threads=self.num_threads):
                    for i in range(self.Nk):
                        self.ph[k2,j,i] = ( self.ph[k2,j,i] +
                            self.a[k2,k1,j,i] * self.qh[k1,j,i] )

        # calculate spectral velocities
        for k in range(self.Nz):
            for j in prange(self.Nl, nogil=True, schedule='static',
                      chunksize=self.chunksize,  
                      num_threads=self.num_threads):
                for i in range(self.Nk):
                    self.uh[k,j,i] = -self._il[j] * self.ph[k,j,i]
                    self.vh[k,j,i] =  self._ik[i] * self.ph[k,j,i]

        # transform to get u and v
        with gil:
            #self.ifft_qh_to_q() # necessary now that timestepping is inside kernel
            self.ifft_uh_to_u()
            self.ifft_vh_to_v()
            
        return
    
    def _do_advection(self):
        self.__do_advection()
    
    cdef void __do_advection(self) nogil:
        ### algorithm
        # uq, vq = (u+Ubg)*q, (v+Vbg)*q
        # uqh, vqh, = fft(uq), fft(vq)
        # tend = kj*uqh + _ilQx*ph + lj*vqh + _ilQy*ph
        
        # the output array: spectal representation of advective tendency
        #cdef np.ndarray tend = np.zeros((self.Nz, self.Nl, self.Nk), dtype=DTYPE_com)

        cdef Py_ssize_t k, j, i

        # multiply to get advective flux in space
        for k in range(self.Nz):
            for j in prange(self.Ny, nogil=True, schedule='static',
                      chunksize=self.chunksize,  
                      num_threads=self.num_threads):
                for i in range(self.Nx):
                    self.uq[k,j,i] = (self.u[k,j,i]+self.Ubg[k]) * self.q[k,j,i]
                    self.vq[k,j,i] = self.v[k,j,i] * self.q[k,j,i]

        # transform to get spectral advective flux
        with gil:
            self.fft_uq_to_uqh()
            self.fft_vq_to_vqh()

        # spectral divergence
        for k in range(self.Nz):
            for j in prange(self.Nl, nogil=True, schedule='static',
                      chunksize=self.chunksize,  
                      num_threads=self.num_threads):
                for i in range(self.Nk):
                    # overwrite the tendency, since the forcing gets called after
                    self.dqhdt[k,j,i] = -( self._ik[i] * self.uqh[k,j,i] +
                                    self._il[j] * self.vqh[k,j,i] +
                                    self._ikQy[k,i] * self.ph[k,j,i] )
        return

    def _do_friction(self):
        self.__do_friction()
    
    cdef void __do_friction(self) nogil:
        """Apply Ekman friction to lower layer tendency"""
        cdef Py_ssize_t k = self.Nz-1
        cdef Py_ssize_t j, i
        if self._rek:
            for j in prange(self.Nl, nogil=True, schedule='static',
                      chunksize=self.chunksize,  
                      num_threads=self.num_threads):
                for i in range(self.Nk):
                    self.dqhdt[k,j,i] = (
                     self.dqhdt[k,j,i] +
                             (self._rek *
                             self._k2l2[j,i] *
                             self.ph[k,j,i]) )
        return                            
                        
    def _forward_timestep(self):
        """Step forward based on tendencies"""
        self.__forward_timestep()

    cdef void __forward_timestep(self) nogil:
   
        #self.dqhdt = self.dqhdt_adv + self.dqhdt_forc
        cdef DTYPE_real_t dt1
        cdef DTYPE_real_t dt2
        cdef DTYPE_real_t dt3
        cdef Py_ssize_t k, j, i
        cdef DTYPE_com_t [:, :, :] qh_new
        with gil:
            qh_new = self.qh.copy()
        
        # Note that Adams-Bashforth is not self-starting
        if self.tc==0:
            # forward euler
            dt1 = self._dt
            dt2 = 0.0
            dt3 = 0.0
        elif self.tc==1:
            # AB2 at step 2
            dt1 = 1.5*self._dt
            dt2 = -0.5*self._dt
            dt3 = 0.0
        else:
            # AB3 from step 3 on
            dt1 = 23./12.*self._dt
            dt2 = -16./12.*self._dt
            dt3 = 5./12.*self._dt
            
        for k in range(self.Nz):
            for j in prange(self.Nl, nogil=True, schedule='static',
                      chunksize=self.chunksize,  
                      num_threads=self.num_threads):
                for i in range(self.Nk):
                    qh_new[k,j,i] = self._filtr[j,i] * (
                        self.qh[k,j,i] +
                        dt1 * self.dqhdt[k,j,i] +
                        dt2 * self.dqhdt_p[k,j,i] +
                        dt3 * self.dqhdt_pp[k,j,i]
                    )
                    self.qh[k,j,i] = qh_new[k,j,i]
                    self.dqhdt_pp[k,j,i] = self.dqhdt_p[k,j,i]
                    self.dqhdt_p[k,j,i] = self.dqhdt[k,j,i]
                    #self.dqhdt[k,j,i] = 0.0
        
        # do FFT of new qh
        with gil:
            self.ifft_qh_to_q() # this destroys qh, need to assign again
        
        for k in range(self.Nz):
            for j in prange(self.Nl, nogil=True, schedule='static',
                      chunksize=self.chunksize,  
                      num_threads=self.num_threads):
                for i in range(self.Nk):
                    self.qh[k,j,i] = qh_new[k,j,i]    

        self.tc += 1
        self.t += self._dt
        return
        
    # attribute aliases: return numpy ndarray views of memory views
    property q:
        def __get__(self):
            return np.asarray(self.q)
    property qh:
        def __get__(self):
            return np.asarray(self.qh)
    property dqhdt:
        def __get__(self):
            return np.asarray(self.dqhdt)
    property dqhdt_p:
        def __get__(self):
            return np.asarray(self.dqhdt_p)
    property dqhdt_pp:
        def __get__(self):
            return np.asarray(self.dqhdt_pp)
    property ph:
        def __get__(self):
            return np.asarray(self.ph)
    property u:
        def __get__(self):
            return np.asarray(self.u)
    property v:
        def __get__(self):
            return np.asarray(self.v)
    property ufull:
        def __get__(self):
            return np.asarray(self.u) + \
                np.asarray(self.Ubg)[:,np.newaxis,np.newaxis]
    property vfull:
        def __get__(self):
            return np.asarray(self.v)
    property uh:
        def __get__(self):
            return np.asarray(self.uh)
    property vh:
        def __get__(self):
            return np.asarray(self.vh)

    property uq:
        def __get__(self):
            return np.asarray(self.uq)
    property vq:
        def __get__(self):
            return np.asarray(self.vq)


# general purpose timestepping routines
# take only complex values, since that what the state variables are
def tendency_forward_euler(DTYPE_real_t dt,
                    np.ndarray[DTYPE_com_t, ndim=3] dqdt):
    """Compute tendency using forward euler timestepping."""
    return dt * dqdt

def tendency_ab2(DTYPE_real_t dt,
                    np.ndarray[DTYPE_com_t, ndim=3] dqdt,
                    np.ndarray[DTYPE_com_t, ndim=3] dqdt_p):
    """Compute tendency using Adams Bashforth 2nd order timestepping."""
    cdef DTYPE_real_t DT1 = 1.5*dt
    cdef DTYPE_real_t DT2 = -0.5*dt
    return DT1 * dqdt + DT2 * dqdt_p

def tendency_ab3(DTYPE_real_t dt,
                    np.ndarray[DTYPE_com_t, ndim=3] dqdt,
                    np.ndarray[DTYPE_com_t, ndim=3] dqdt_p,
                    np.ndarray[DTYPE_com_t, ndim=3] dqdt_pp):
    """Compute tendency using Adams Bashforth 3nd order timestepping."""
    cdef DTYPE_real_t DT1 = 23/12.*dt
    cdef DTYPE_real_t DT2 = -16/12.*dt
    cdef DTYPE_real_t DT3 = 5/12.*dt
    return DT1 * dqdt + DT2 * dqdt_p + DT3 * dqdt_pp 
    















__version__='0.1.3'
from .model import Model
from .qg_model import QGModel
from .bt_model import BTModel
from .sqg_model import SQGModel
from .layered_model import LayeredModel
from .particles import LagrangianParticleArray2D, GriddedLagrangianParticleArray2D



import numpy as np
from particles import LagrangianParticleArray2D
r_twopi = (2*np.pi)**-1

class PointVortexArray2D(LagrangianParticleArray2D):
    """Keeps track of a set of point vortices.""" 
    
    def __init__(self, x0, y0, s0, **kwargs):
        
        # strength
        self.s = s0
        
        LagrangianParticleArray2D.__init__(self,
            x0, y0, **kwargs)
            
        # need to remember previous position
        self.xprev = self.x.copy()
        self.yprev = self.y.copy()
        
    def step_forward_vortices(self, dt):       
        # create interpolation functions which return u and v
        uv0fun = (lambda x, y : self.calc_uv(x, y, prev=True))
        uv1fun = (lambda x, y : self.calc_uv(x, y))
     
        dx, dy = self.rk4_integrate(self.x, self.y, uv0fun, uv1fun, dt)

        self.xprev = self.x.copy()
        self.yprev = self.y.copy()
        self.x = self.wrap_x(self.x + dx)
        self.y = self.wrap_y(self.y + dy)

    def calc_uv(self, x, y, prev=False):
        """Calculate velocity at x and y points due to vortex velocity field.
        Assumes x and y are vortex positions and are ordered the same as
        x0 and y0. The ordering is used to neglect to vortex self interaction."""
        assert len(x) == self.N
        assert len(y) == self.N
        u = np.zeros(self.N, self.x.dtype)
        v = np.zeros(self.N, self.y.dtype)
        for n in xrange(self.N):
            # don't include self interaction
            if prev:
                x0 = self.xprev[np.r_[:n,n+1:self.N]]
                y0 = self.yprev[np.r_[:n,n+1:self.N]]
            else:
                x0 = self.x[np.r_[:n,n+1:self.N]]
                y0 = self.y[np.r_[:n,n+1:self.N]]
            s0 = self.s[np.r_[:n,n+1:self.N]]
            u0, v0 = self.uv_at_xy(x[n], y[n], x0, y0, s0)
            u[n] = u0.sum()
            v[n] = v0.sum()
        return u, v
        
    def uv_at_xy(self, x, y, x0, y0, s0):
        """Returns two arrays of u, v"""
        dx, dy = self.distance(x0, y0, x, y)
        #print 'dx, dy:', dx, dy
        rr2 = (dx**2 + dy**2)**-1
        u = - s0 * dy * r_twopi * rr2
        v = s0 * dx * r_twopi * rr2
        #print 'u, v', u, v
        return u, v




from setuptools import setup, Extension
from Cython.Build import cythonize
import warnings
import numpy as np
import os
import tempfile, subprocess, shutil

VERSION='0.2.0'

DISTNAME='pyqg'
URL='http://github.com/pyqg/pyqg'
# how can we make download_url automatically get the right version?
DOWNLOAD_URL='https://github.com/pyqg/pyqg/tarball/v%s' % VERSION
AUTHOR='pyqg team'
AUTHOR_EMAIL='pyqg-dev@googlegroups.com'
LICENSE='MIT'

DESCRIPTION='python quasigeostrophic model'
LONG_DESCRIPTION="""
pyqg is a python solver for quasigeostrophic systems. Quasigeostophic
equations are an approximation to the full fluid equations of motion in
the limit of strong rotation and stratitifcation and are most applicable
to geophysical fluid dynamics problems.

Students and researchers in ocean and atmospheric dynamics are the intended
audience of pyqg. The model is simple enough to be used by students new to
the field yet powerful enough for research. We strive for clear documentation
and thorough testing.

pyqg supports a variety of different configurations using the same
computational kernel. The different configurations are evolving and are
described in detail in the documentation. The kernel, implement in cython,
uses a pseudo-spectral method which is heavily dependent of the fast Fourier
transform. For this reason, pyqg depends on pyfftw and the FFTW Fourier
Transform library. The kernel is multi-threaded but does not support mpi.
Optimal performance will be achieved on a single system with many cores.

Links
-----

- HTML documentation: http://pyqg.readthedocs.org
- Issue tracker: http://github.com/pyqg/pyqg/issues
- Source code: http://github.com/pyqg/pyqg
"""

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Physics',
    'Topic :: Scientific/Engineering :: Atmospheric Science'
]


### Dependency section ###
install_requires = [
    'cython',
    'numpy'
]

# This hack tells cython whether pyfftw is present
use_pyfftw_file = 'pyqg/.compile_time_use_pyfftw.pxi'
with open(use_pyfftw_file, 'wb') as f:
    try:
        import pyfftw
        f.write(b'DEF PYQG_USE_PYFFTW = 1')
    except ImportError:
        f.write(b'DEF PYQG_USE_PYFFTW = 0')
        warnings.warn('Could not import pyfftw. Model will be slow.')

# check for openmp following
# http://stackoverflow.com/questions/16549893/programatically-testing-for-openmp-support-from-a-python-setup-script
# see http://openmp.org/wp/openmp-compilers/
omp_test = \
br"""
#include <omp.h>
#include <stdio.h>
int main() {
#pragma omp parallel
printf("Hello from thread %d, nthreads %d\n", omp_get_thread_num(), omp_get_num_threads());
}
"""

# python 3 needs rb

def check_for_openmp():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    filename = r'test.c'
    try:
        cc = os.environ['CC']
    except KeyError:
        cc = 'gcc'
    with open(filename, 'wb', 0) as file:
        file.write(omp_test)
    with open(os.devnull, 'wb') as fnull:
        result = subprocess.call([cc, '-fopenmp', filename],
                                 stdout=fnull, stderr=fnull)
    print('check_for_openmp() result: ', result)
    os.chdir(curdir)
    #clean up
    shutil.rmtree(tmpdir)

    return result==0

extra_compile_args = []
extra_link_args = []

use_openmp = True
if check_for_openmp() and use_openmp:
    extra_compile_args.append('-fopenmp')
    extra_link_args.append('-fopenmp')
else:
    warnings.warn('Could not link with openmp. Model will be slow.')

# reathedocs can't and shouldn't build pyfftw
# apparently setup.py overrides docs/requirements.txt
#on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
#if on_rtd:
#    install_requires.remove('pyfftw')

tests_require = ['pytest']

def readme():
    with open('README.md') as f:
        return f.read()

ext_module = Extension(
    "pyqg.kernel",
    ["pyqg/kernel.pyx"],
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
)

setup(name=DISTNAME,
      version=VERSION,
      description=DESCRIPTION,
      classifiers=CLASSIFIERS,
      long_description=LONG_DESCRIPTION,
      url=URL,
      download_url=DOWNLOAD_URL,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      packages=['pyqg'],
      install_requires=install_requires,
      ext_modules = cythonize(ext_module),
      include_dirs = [np.get_include()],
      tests_require = tests_require,
      test_suite = 'nose.collector',
      zip_safe=False)



import warnings
import numpy as np
#from scipy.interpolate import RectBivariateSpline
#from regulargrid.cartesiangrid import CartesianGrid
#from scipy.ndimage import map_coordinates
# works better with mock on readthedocs

try:
    import scipy.ndimage
except ImportError:
    warnings.warn('Failed to import scipy.ndimage. '
           'Gridded interpolation will not work',
	   ImportWarning)

class LagrangianParticleArray2D(object):
    """A class for keeping track of a set of lagrangian particles
    in two-dimensional space. Tries to be fast.
    """
    
    def __init__(self, x0, y0,
                       periodic_in_x=False,
                       periodic_in_y=False,
                       xmin=-np.inf, xmax=np.inf,
                       ymin=-np.inf, ymax=np.inf,
                       particle_dtype='f8'):
        """
        Parameters
        ---------- 
        
        x0, y0 : array-like
            Two arrays (same size) representing the particle initial
            positions.
        periodic_in_x : bool
            Whether the domain wraps in the x direction.
        periodic_in_y : bool
            Whether the domain 'wraps' in the y direction.
        xmin, xmax : numbers
            Maximum and minimum values of x coordinate
        ymin, ymax : numbers
            Maximum and minimum values of y coordinate
        particle_dtype : dtype
            Data type to use for particles
        """
        
        self.x = np.array(x0, dtype=np.dtype(particle_dtype)).ravel()
        self.y = np.array(y0, dtype=np.dtype(particle_dtype)).ravel()
        
        assert self.x.shape == self.y.shape
        self.N = len(self.x)
        
        # check that the particles are within the specified boundaries
        assert np.all(self.x >= xmin) and np.all(self.x <= xmax)
        assert np.all(self.y >= ymin) and np.all(self.y <= ymax)
        
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.pix = periodic_in_x
        self.piy = periodic_in_y
                        
        self.Lx = self.xmax - self.xmin
        self.Ly = self.ymax - self.ymin
                 
    def step_forward_with_function(self, uv0fun, uv1fun, dt):
        """Advance particles using a function to determine u and v.
        
        Parameters
        ----------
        uv0fun : function
            Called like ``uv0fun(x,y)``. Should return the velocity field
            u, v at time t.
        uv1fun(x,y) : function
            Called like ``uv1fun(x,y)``. Should return the velocity field
            u, v at time t + dt.
        dt : number
            Timestep."""
            
        dx, dy = self._rk4_integrate(self.x, self.y, uv0fun, uv1fun, dt)
        self.x = self._wrap_x(self.x + dx)
        self.y = self._wrap_y(self.y + dy)
    
    def _rk4_integrate(self, x, y, uv0fun, uv1fun, dt):
        """Integrates positions x, y using velocity functions
           uv0fun, uv1fun. Returns dx and dy, the displacements."""
        u0, v0 = uv0fun(x, y)
        k1u = dt*u0
        k1v = dt*v0
        x11 = self._wrap_x(x + 0.5*k1u)
        y11 = self._wrap_y(y + 0.5*k1v)
        u11, v11 = uv1fun(x11, y11)
        k2u = dt*u11
        k2v = dt*v11
        x12 = self._wrap_x(x + 0.5*k2u)
        y12 = self._wrap_y(y + 0.5*k2v)
        u12, v12 = uv1fun(x12, y12)
        k3u = dt*u12
        k3v = dt*v12
        x13 = self._wrap_x(x + k3u)
        y13 = self._wrap_y(y + k3v)
        u13, v13 = uv1fun(x13, y13)
        k4u = dt*u13
        k4v = dt*v13
        
        # update
        dx = 6**-1*(k1u + 2*k2u + 2*k3u + k4u)
        dy = 6**-1*(k1v + 2*k2v + 2*k3v + k4v)
        return dx, dy
        
    def _wrap_x(self, x):
        # wrap positions
        if self.pix:
            return np.mod(x-self.xmin, self.Lx) + self.xmin
        else:
            return x
    
    def _wrap_y(self, y):
        # wrap y position
        if self.piy:
            return np.mod(y-self.ymin, self.Ly) + self.ymin
        else:
            return y
            
    def _distance(self, x0, y0, x1, y1):
        """Utitlity function to compute distance between points."""
        dx = x1-x0
        dy = y1-y0
        # roll displacements across the borders
        if self.pix:
            dx[ dx > self.Lx/2 ] -= self.Lx
            dx[ dx < -self.Lx/2 ] += self.Lx
        if self.piy:
            dy[ dy > self.Ly/2 ] -= self.Ly
            dy[ dy < -self.Ly/2 ] += self.Ly
        return dx, dy
        
        
class GriddedLagrangianParticleArray2D(LagrangianParticleArray2D):
    """Lagrangian particles with velocities given on a regular cartesian grid.
    """
    
    def __init__(self, x0, y0, Nx, Ny, grid_type='A', **kwargs):
        """
        Parameters
        ---------- 
        
        x0, y0 : array-like
            Two arrays (same size) representing the particle initial
            positions.
        Nx, Ny: int
            Number of grid points in the x and y directions
        grid_type: {'A'}
            Arakawa grid type specifying velocity positions.
        """
        
        super(GriddedLagrangianParticleArray2D, self).__init__(x0, y0, **kwargs)
        self.Nx = Nx
        self.Ny = Ny
        
        if grid_type != 'A':
            raise ValueError('Only A grid velocities supported at this time.')
            
        if not (self.pix and self.piy):
            raise ValueError(
              'Interpolation only works with doubly periodic grids at this time.')
            
        # figure out grid geometry, assuming velocities are located a cell centers
        
    # def xy_to_ij(self, x, y):
    #     """Convert spatial coords x, y to grid coords i, j"""
    #     i = x/self.Lx*self.Nx - 0.5
    #     j = y/self.Ly*self.Ny - 0.5
    #     return i, j
    #
    # def ij_to_xy(self, i, j):
    #     """Convert grid coords i, j to spatial coords x, y"""
    #     x = (i + 0.5)*self.Lx/self.Nx
    #     y = (j + 0.5)*self.Ly/self.Ny
    
    def interpolate_gridded_scalar(self, x, y, c, order=1, pad=1, offset=0):
        """Interpolate gridded scalar C to points x,y.
        
        Parameters
        ----------
        x, y : array-like
            Points at which to interpolate
        c : array-like
            The scalar, assumed to be defined on the grid.
        order : int
            Order of interpolation
        pad : int
            Number of pad cells added
        offset : int
            ??? 
            
        Returns
        -------
        ci : array-like
            The interpolated scalar
        """
        
        ## no longer necessary because we accept pre-padded arrays
        # assert c.shape == (self.Ny, self.Nx), 'Shape of c needs to be (Ny,Nx)'
        
        # first pad the array to deal with the boundaries
        # (map_coordinates can't seem to deal with this by itself)
        # pad twice so cubic interpolation can be used
        if pad > 0:
            cp = self._pad_field(c, pad=pad)
        else:
            cp = c
        # now the shape is Nx+2, Nx+2
        i = (x - self.xmin)/self.Lx*self.Nx + pad + offset - 0.5
        j = (y - self.ymin)/self.Ly*self.Ny + pad + offset - 0.5
        
        # for some reason this still does not work with high precision near the boundaries
        return scipy.ndimage.map_coordinates(cp, [j,i],
                mode='constant', order=order, cval=np.nan)
    
    def _pad_field(self, c, pad=5):
        return np.pad(c, ((pad,pad),(pad,pad)), mode='wrap')
           
    def step_forward_with_gridded_uv(self, U0, V0, U1, V1, dt, order=1):       
        """Advance particles using a gridded velocity field. Because of the
        Runga-Kutta timestepping, we need two velocity fields at different
        times.
        
        Parameters
        ----------
        U0, V0 : array-like
            Gridded velocity fields at time t - dt.
        U1, V1 : array-like
            Gridded velocity fields at time t.
        dt : number
            Timestep.
        order : int
            Order of interpolation.
        """
        # create interpolation functions which return u and v
        
        # pre-pad arrays so it only has to be done once
        # for linear interpolation (default order=1), only one pad is necessary
        pad = order
        [U0p, V0p, U1p, V1p] = [self._pad_field(c, pad=pad) for c in [U0, V0, U1, V1]]
        
        # pad u and v as necessary
        uv0fun = (lambda x, y : 
                  (self.interpolate_gridded_scalar(x, y, U0p,
                           pad=0, order=order, offset=pad),
                   self.interpolate_gridded_scalar(x, y, V0p,
                           pad=0, order=order, offset=pad)))
        uv1fun = (lambda x, y :  
                  (self.interpolate_gridded_scalar(x, y, U1p, pad=0,
                           order=order, offset=pad),
                   self.interpolate_gridded_scalar(x, y, V1p, pad=0,
                           order=order, offset=pad)))
        
        self.step_forward_with_function(uv0fun, uv1fun, dt)
        #dx, dy = self.rk4_integrate(self.x, self.y, uv0fun, uv1fun, dt)
        #self.x = self.wrap_x(self.x + dx)
        #self.y = self.wrap_y(self.y + dy)
        
        
        
        
        
        
        
        
        
        
        



from __future__ import print_function
import numpy as np
from numpy import pi
import logging
from .kernel import PseudoSpectralKernel, tendency_forward_euler, tendency_ab2, tendency_ab3
try:
    import mkl
    np.use_fastnumpy = True
except ImportError:
    pass

try:
    import pyfftw
    pyfftw.interfaces.cache.enable()
except ImportError:
    pass

class Model(PseudoSpectralKernel):
    """A generic pseudo-spectral inversion model.

    Attributes
    ----------
    q : real array
        Potential vorticity in real space
    qh : complex array
        Potential vorticity in spectral space
    ph : complex array
        Streamfunction in spectral space
    u, v : real arrays
        Velocity anomaly components in real space
    ufull, vfull : real arrays
        Full velocity components in real space
    uh, vh : complex arrays
        Velocity anomaly components in spectral space
    nx, ny : int
        Number of grid points in the x and y directions
    L, W : float
        Domain length in x and y directions
    rek : float
        Linear drag in lower layer
    filterfac : float
        Amplitdue of the spectral spherical filter
    dt : float
        Numerical timstep
    twrite : int
        Interval for cfl writeout (units: number of timesteps)
    tmax : float
        Total time of integration (units: model time)
    tavestart : float
        Start time for averaging (units: model time)
    tsnapstart : float
        Start time for snapshot writeout (units: model time)
    taveint : float
        Time interval for accumulation of diagnostic averages.
        (units: model time)
    tsnapint : float
        Time interval for snapshots (units: model time)
    ntd : int
        Number of threads to use. Should not exceed the number of cores on
        your machine.
    pmodes : real array
        Vertical pressure modes (unitless)
    radii :  real array
        Deformation radii  (units: model length)
    """

    def __init__(
        self,
        # grid size parameters
        nx=64,                     # grid resolution
        ny=None,
        L=1e6,                     # domain size is L [m]
        W=None,
        # timestepping parameters
        dt=7200.,                   # numerical timestep
        twrite=1000.,               # interval for cfl and ke writeout (in timesteps)
        tmax=1576800000.,           # total time of integration
        tavestart=315360000.,       # start time for averaging
        taveint=86400.,             # time interval used for summation in longterm average in seconds
        useAB2=False,               # use second order Adams Bashforth timestepping instead of 3rd
        # friction parameters
        rek=5.787e-7,               # linear drag in lower layer
        filterfac=23.6,             # the factor for use in the exponential filter
        # constants
        f = None,                   # coriolis parameter (not necessary for two-layer model
                                    #  if deformation radius is provided)
        g= 9.81,                    # acceleration due to gravity
        # diagnostics parameters
        diagnostics_list='all',     # which diagnostics to output
        # fft parameters
        # removed because fftw is now manditory
        #use_fftw = False,               # fftw flag
        #teststyle = False,            # use fftw with "estimate" planner to get reproducibility
        ntd = 1,                       # number of threads to use in fftw computations
        log_level = 1,                 # logger level: from 0 for quiet (no log) to 4 for verbose
                                       #     logger (see  https://docs.python.org/2/library/logging.html)
        logfile = None,                # logfile; None prints to screen
        ):
        """
        .. note:: All of the test cases use ``nx==ny``. Expect bugs if you choose
                  these parameters to be different.
        .. note:: All time intervals will be rounded to nearest `dt` interval.

        Parameters
        ----------
        nx : int
            Number of grid points in the x direction.
        ny : int
            Number of grid points in the y direction (default: nx).
        L : number
            Domain length in x direction. Units: meters.
        W :
            Domain width in y direction. Units: meters (default: L).
        rek : number
            linear drag in lower layer. Units: seconds :sup:`-1`.
        filterfac : number
            amplitdue of the spectral spherical filter (originally 18.4, later
            changed to 23.6).
        dt : number
            Numerical timstep. Units: seconds.
        twrite : int
            Interval for cfl writeout. Units: number of timesteps.
        tmax : number
            Total time of integration. Units: seconds.
        tavestart : number
            Start time for averaging. Units: seconds.
        tsnapstart : number
            Start time for snapshot writeout. Units: seconds.
        taveint : number
            Time interval for accumulation of diagnostic averages.
            Units: seconds. (For performance purposes, averaging does not have to
            occur every timestep)
        tsnapint : number
            Time interval for snapshots. Units: seconds.
        ntd : int
            Number of threads to use. Should not exceed the number of cores on
            your machine.
        """

        if ny is None: ny = nx
        if W is None: W = L

        # put all the parameters into the object
        # grid
        self.nx = nx
        self.ny = ny
        self.L = L
        self.W = W

        # timestepping
        self.dt = dt
        self.twrite = twrite
        self.tmax = tmax
        self.tavestart = tavestart
        self.taveint = taveint
        self.logfile = logfile
        self.log_level = log_level
        self.useAB2 = useAB2
        # fft
        #self.use_fftw = use_fftw
        #self.teststyle= teststyle
        self.ntd = ntd

        # friction
        self.rek = rek
        self.filterfac = filterfac

        # constants
        self.g = g
        if f:
            self.f = f
            self.f2 = f**2


        self._initialize_logger()
        self._initialize_grid()
        self._initialize_background()
        self._initialize_forcing()
        self._initialize_filter()
        self._initialize_inversion_matrix()
        self._initialize_time()

        # call the underlying cython kernel
        self._initialize_kernel()

        self._initialize_diagnostics(diagnostics_list)

    def run_with_snapshots(self, tsnapstart=0., tsnapint=432000.):
        """Run the model forward, yielding to user code at specified intervals.

        Parameters
        ----------

        tsnapstart : int
            The timestep at which to begin yielding.
        tstapint : int
            The interval at which to yield.
        """

        tsnapints = np.ceil(tsnapint/self.dt)

        while(self.t < self.tmax):
            self._step_forward()
            if self.t>=tsnapstart and (self.tc%tsnapints)==0:
                yield self.t
        return

    def run(self):
        """Run the model forward without stopping until the end."""
        while(self.t < self.tmax):
            self._step_forward()


    def vertical_modes(self):
        """ Calculate standard vertical modes. Simply
            the eigenvectors of the stretching matrix S """

        evals,evecs = np.linalg.eig(-self.S)

        asort = evals.argsort()

        # deformation wavenumbers and radii
        self.kdi2 = evals[asort]
        self.radii = np.zeros_like(self.kdi2)
        self.radii[0] = self.g*self.H/np.abs(self.f) # barotropic def. radius
        self.radii[1:] = 1./np.sqrt(self.kdi2[1:])

        # eigenstructure
        self.pmodes = evecs[:,asort]

        # normalize to have unit L2-norm
        Ai = (self.H / (self.Hi[:,np.newaxis]*(self.pmodes**2)).sum(axis=0))**0.5
        self.pmodes = Ai[np.newaxis,:]*self.pmodes

    def modal_projection(self,p,forward=True):
        """ Performs a field p into modal amplitudes pn
                using the basis [pmodes]. The inverse
                transform calculates p from pn"""

        if forward:
            pt = np.linalg.solve(self.pmodes[np.newaxis,np.newaxis],p.T).T
        else:
            pt = np.einsum("ik,k...->i...",self.pmodes,p)

        return pt

    def stability_analysis(self,bottom_friction=False):
        """ Performs the baroclinic linear instability analysis given
                given the base state velocity :math: `(U, V)` and
                the stretching matrix  :math: `S`:

        .. math:: A \Phi = \omega B \Phi,

        where

        .. math:: A = B (U k + V l) + I (k Q_y - l Q_x) +
                        1j \delta_{N N} r_{ek} I  \kappa^2

        where :math:`\delta_{N N} = [0,0,\dots,0,1] ,`

        and

        .. math:: B =  S - I \kappa^2 .


        The eigenstructure is

        .. math:: \Phi

        and the eigenvalue is

        .. math:: `\omega`

        The growth rate is Im\ :math:`\{\omega\}`.


        Parameters
        ----------
        bottom_friction: optional inclusion linear bottom drag
                         in the linear stability calculation
                         (default is False, as if :math: `r_{ek} = 0`)

        Returns
        -------
        omega: complex array
             The eigenvalues with largest complex part (units: inverse model time)
        phi: complex array
             The eigenvectors associated associated with \omega (unitless)

        """

        omega = np.zeros_like(self.wv)+0.j
        phi = np.zeros_like(self.qh)

        I = np.eye(self.nz)

        L2 = self.S[:,:,np.newaxis,np.newaxis] - self.wv2*I[:,:,np.newaxis,np.newaxis]

        Q =  I[:,:,np.newaxis,np.newaxis]*(self.ikQy - self.ilQx).imag

        Uk =(self.Ubg*I)[:,:,np.newaxis,np.newaxis]*self.k
        Vl =(self.Vbg*I)[:,:,np.newaxis,np.newaxis]*self.l
        L3 = np.einsum('ij...,jk...->ik...',L2,Uk+Vl) + 0j

        if bottom_friction:
            L3[-1,-1,:,:] += 1j*self.rek*self.wv2

        L4 = self.a.T

        M = np.einsum('...ij,...jk->...ik',L4,(L3+Q).T)

        evals,evecs = np.linalg.eig(M)
        evals, evecs = evals.T, evecs.T

        # sorting things this way proved way
        #  more faster than using numpy's argsort() !
        imax = evals.imag.argmax(axis=0)
        for i in range(self.nl):
            for j in range(self.nk):
                omega[i,j] = evals[imax[i,j],i,j]
                phi[:,i,j] = evecs[imax[i,j],:,i,j]

        return omega, phi

    ### PRIVATE METHODS - not meant to be called by user ###

    def _step_forward(self):

        self._invert()
        # find streamfunction from pv

        self._do_advection()
        # use streamfunction to calculate advection tendency

        self._do_friction()
        # apply friction

        self._do_external_forcing()
        # apply external forcing

        self._calc_diagnostics()
        # do what has to be done with diagnostics

        self._forward_timestep()
        # apply tendencies to step the model forward
        # (filter gets called here)

        # the basic steps are
        self._print_status()

    def _initialize_time(self):
        """Set up timestep stuff"""
        #self.t=0        # actual time
        #self.tc=0       # timestep number
        self.taveints = np.ceil(self.taveint/self.dt)

    ### initialization routines, only called once at the beginning ###
    def _initialize_grid(self):
        """Set up spatial and spectral grids and related constants"""
        self.x,self.y = np.meshgrid(
            np.arange(0.5,self.nx,1.)/self.nx*self.L,
            np.arange(0.5,self.ny,1.)/self.ny*self.W )

        # Notice: at xi=1 U=beta*rd^2 = c for xi>1 => U>c
        # wavenumber one (equals to dkx/dky)
        self.dk = 2.*pi/self.L
        self.dl = 2.*pi/self.W

        # wavenumber grids
        self.nl = self.ny
        self.nk = int(self.nx/2+1)
        self.ll = self.dl*np.append( np.arange(0.,self.nx/2),
            np.arange(-self.nx/2,0.) )
        self.kk = self.dk*np.arange(0.,self.nk)

        self.k, self.l = np.meshgrid(self.kk, self.ll)
        self.ik = 1j*self.k
        self.il = 1j*self.l
        # physical grid spacing
        self.dx = self.L / self.nx
        self.dy = self.W / self.ny

        # constant for spectral normalizations
        self.M = self.nx*self.ny

        # isotropic wavenumber^2 grid
        # the inversion is not defined at kappa = 0
        self.wv2 = self.k**2 + self.l**2
        self.wv = np.sqrt( self.wv2 )

        iwv2 = self.wv2 != 0.
        self.wv2i = np.zeros_like(self.wv2)
        self.wv2i[iwv2] = self.wv2[iwv2]**-1

    def _initialize_background(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_inversion_matrix(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_forcing(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_filter(self):
        """Set up frictional filter."""
        # this defines the spectral filter (following Arbic and Flierl, 2003)
        cphi=0.65*pi
        wvx=np.sqrt((self.k*self.dx)**2.+(self.l*self.dy)**2.)
        self.filtr = np.exp(-self.filterfac*(wvx-cphi)**4.)
        self.filtr[wvx<=cphi] = 1.

    def _filter(self, q):
        return self.filtr * q

    def _do_external_forcing(self):
        pass

    def _initialize_kernel(self):
        #super(spectral_kernel.PseudoSpectralKernel, self).__init__(
        self._kernel_init(
            self.nz, self.ny, self.nx,
            self.a, self.kk, self.ll,
            self.Ubg, self.Qy,
            self.filtr,
            dt=self.dt,
            rek=self.rek,
            fftw_num_threads=self.ntd
        )

        # still need to initialize a few state variables here, outside kernel
        # this is sloppy
        #self.dqhdt_forc = np.zeros_like(self.qh)
        #self.dqhdt_p = np.zeros_like(self.qh)
        #self.dqhdt_pp = np.zeros_like(self.qh)

        self.logger.info(' Kernel initialized')


    # logger
    def _initialize_logger(self):

        self.logger = logging.getLogger(__name__)


        if not (self.logfile is None):
            fhandler = logging.FileHandler(filename=self.logfile, mode='w')
        else:
            fhandler = logging.StreamHandler()

        formatter = logging.Formatter('%(levelname)s: %(message)s')

        fhandler.setFormatter(formatter)

        if not self.logger.handlers:
            self.logger.addHandler(fhandler)

        self.logger.setLevel(self.log_level*10)

        # this prevents the logger to propagate into the ipython notebook log
        self.logger.propagate = False

        self.logger.info(' Logger initialized')


    # compute advection in grid space (returns qdot in fourier space)
    # *** don't remove! needed for diagnostics (but not forward model) ***
    def _advect(self, q, u, v):
        """Given real inputs q, u, v, returns the advective tendency for
        q in spectal space."""
        uq = u*q
        vq = v*q
        # this is a hack, since fft now requires input to have shape (nz,ny,nx)
        # it does an extra unnecessary fft
        is_2d = (uq.ndim==2)
        if is_2d:
            uq = np.tile(uq[np.newaxis,:,:], (self.nz,1,1))
            vq = np.tile(vq[np.newaxis,:,:], (self.nz,1,1))
        tend = self.ik*self.fft(uq) + self.il*self.fft(vq)
        if is_2d:
            return tend[0]
        else:
            return tend

    # def _filter(self, q):
    #     """Apply filter to field q."""
    #     return q

    def _print_status(self):
        """Output some basic stats."""
        if (self.log_level) and ((self.tc % self.twrite)==0):
            self.ke = self._calc_ke()
            self.cfl = self._calc_cfl()
            #print 't=%16d, tc=%10d: cfl=%5.6f, ke=%9.9f' % (
            #       self.t, self.tc, cfl, ke)
            self.logger.info('Step: %i, Time: %3.2e, KE: %3.2e, CFL: %4.3f'
                    , self.tc,self.t,self.ke,self.cfl )

            assert self.cfl<1., self.logger.error('CFL condition violated')


    def _calc_diagnostics(self):
        # here is where we calculate diagnostics
        if (self.t>=self.dt) and (self.t>=self.tavestart) and (self.tc%self.taveints==0):
            self._increment_diagnostics()

    # def _forward_timestep(self):
    #     """Step forward based on tendencies"""
    #
    #     #self.dqhdt = self.dqhdt_adv + self.dqhdt_forc
    #
    #     # Note that Adams-Bashforth is not self-starting
    #     if self.tc==0:
    #         # forward Euler at the first step
    #         qtend = tendency_forward_euler(self.dt, self.dqhdt)
    #     elif (self.tc==1) or (self.useAB2):
    #         # AB2 at step 2
    #         qtend = tendency_ab2(self.dt, self.dqhdt, self.dqhdt_p)
    #     else:
    #         # AB3 from step 3 on
    #         qtend = tendency_ab3(self.dt, self.dqhdt,
    #                     self.dqhdt_p, self.dqhdt_pp)
    #
    #     # add tendency and filter
    #     self.set_qh(self._filter(self.qh + qtend))
    #
    #     # remember previous tendencies
    #     self.dqhdt_pp[:] = self.dqhdt_p.copy()
    #     self.dqhdt_p[:] = self.dqhdt.copy()
    #     #self.dqhdt[:] = 0.
    #
    #     # augment timestep and current time
    #     self.tc += 1
    #     self.t += self.dt

    ### All the diagnostic stuff follows. ###
    def _calc_cfl(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    # this is stuff the Cesar added

    # if self.tc==0:
    #     assert self.calc_cfl()<1., " *** time-step too large "
    #     # initialize ke and time arrays
    #     self.ke = np.array([self.calc_ke()])
    #     self.eddy_time = np.array([self.calc_eddy_time()])
    #     self.time = np.array([0.])


    def _calc_ke(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_diagnostics(self, diagnostics_list):
        # Initialization for diagnotics
        self.diagnostics = dict()

        self._initialize_core_diagnostics()
        self._initialize_model_diagnostics()

        if diagnostics_list == 'all':
            pass # by default, all diagnostics are active
        elif diagnostics_list == 'none':
            self.set_active_diagnostics([])
        else:
            self.set_active_diagnostics(diagnostics_list)

    def _initialize_core_diagnostics(self):
        """Diagnostics common to all models."""
        self.add_diagnostic('Ensspec',
            description='enstrophy spectrum',
            function= (lambda self: np.abs(self.qh)**2/self.M**2)
        )

        self.add_diagnostic('KEspec',
            description=' kinetic energy spectrum',
            function=(lambda self: self.wv2*np.abs(self.ph)**2/self.M**2)
        )      # factor of 2 to account for the fact that we have only half of
               #    the Fourier coefficients.

        self.add_diagnostic('q',
            description='QGPV',
            function= (lambda self: self.q)
        )

        self.add_diagnostic('EKEdiss',
            description='total energy dissipation by bottom drag',
            function= (lambda self: self.Hi[-1]/self.H*self.rek*(self.v[-1]**2 + self.u[-1]**2).mean())
        )

        self.add_diagnostic('EKE',
            description='mean eddy kinetic energy',
            function= (lambda self: 0.5*(self.v**2 + self.u**2).mean(axis=-1).mean(axis=-1))
        )

    def _calc_derived_fields(self):
        """Should be implemented by subclass."""
        pass

    def _initialize_model_diagnostics(self):
        """Should be implemented by subclass."""
        pass

    def _set_active_diagnostics(self, diagnostics_list):
        for d in self.diagnostics:
            self.diagnostics[d]['active'] == (d in diagnostics_list)

    def add_diagnostic(self, diag_name, description=None, units=None, function=None):
        # create a new diagnostic dict and add it to the object array

        # make sure the function is callable
        assert hasattr(function, '__call__')

        # make sure the name is valid
        assert isinstance(diag_name, str)

        # by default, diagnostic is active
        self.diagnostics[diag_name] = {
           'description': description,
           'units': units,
           'active': True,
           'count': 0,
           'function': function, }

    def describe_diagnostics(self):
        """Print a human-readable summary of the available diagnostics."""
        diag_names = self.diagnostics.keys()
        diag_names.sort()
        print('NAME               | DESCRIPTION')
        print(80*'-')
        for k in diag_names:
            d = self.diagnostics[k]
            print('{:<10} | {:<54}').format(
                 *(k,  d['description']))

    def _increment_diagnostics(self):
        # compute intermediate quantities needed for some diagnostics

        self._calc_derived_fields()

        for dname in self.diagnostics:
            if self.diagnostics[dname]['active']:
                res = self.diagnostics[dname]['function'](self)
                if self.diagnostics[dname]['count']==0:
                    self.diagnostics[dname]['value'] = res
                else:
                    self.diagnostics[dname]['value'] += res
                self.diagnostics[dname]['count'] += 1

    def get_diagnostic(self, dname):
        return (self.diagnostics[dname]['value'] /
                self.diagnostics[dname]['count'])

    def spec_var(self, ph):
        """ compute variance of p from Fourier coefficients ph """
        var_dens = 2. * np.abs(ph)**2 / self.M**2
        # only half of coefs [0] and [nx/2+1] due to symmetry in real fft2
        var_dens[...,0] = var_dens[...,0]/2.
        var_dens[...,-1] = var_dens[...,-1]/2.
        return var_dens.sum()



from __future__ import print_function
import numpy as np
from numpy import pi
from . import model


class BTModel(model.Model):
    r"""Single-layer (barotropic) quasigeostrophic model.
    This class can represent both pure two-dimensional flow
    and also single reduced-gravity layers with deformation
    radius ``rd``.

    The equivalent-barotropic quasigeostrophic evolution equations is

    .. math::

       \partial_t q + J(\psi, q ) + \beta \psi_x = \text{ssd}

    The potential vorticity anomaly is

    .. math::

       q = \nabla^2 \psi - \kappa_d^2 \psi

    """

    def __init__(self, beta=0.,  rd=0., H=1., U=0., **kwargs):
        """
        Parameters
        ----------

        beta : number, optional
            Gradient of coriolis parameter. Units: meters :sup:`-1`
            seconds :sup:`-1`
        rd : number, optional
            Deformation radius. Units: meters.
        U : number, optional
            Upper layer flow. Units: meters.
        """

        self.beta = beta
        self.rd = rd
        self.H = H
        self.Hi = np.array(H)[np.newaxis,...]
        self.U = U

        self.nz = 1

        # deformation wavenumber
        if rd:
            self.kd2 = rd**-2
        else:
            self.kd2 = 0.

        super(BTModel, self).__init__(**kwargs)

        # initial conditions: (PV anomalies)
        self.set_q(1e-3*np.random.rand(1,self.ny,self.nx))

    def _initialize_background(self):
        """Set up background state (zonal flow and PV gradients)."""

        # the meridional PV gradients in each layer
        self.Qy = np.asarray(self.beta)[np.newaxis, ...]

        # background vel.
        self.set_U(self.U)

        # complex versions, multiplied by k, speeds up computations to pre-compute
        self.ikQy = self.Qy * 1j * self.k

        self.ilQx = 0.

    def _initialize_inversion_matrix(self):
        """ the inversion """
        # The bt model is diagonal. The inversion is simply qh = -kappa**2 ph
        self.a = -(self.wv2i+self.kd2)[np.newaxis, np.newaxis, :, :]

    def _initialize_forcing(self):
        pass

    # def _initialize_forcing(self):
    #     """Set up frictional filter."""
    #     # this defines the spectral filter (following Arbic and Flierl, 2003)
    #     cphi=0.65*pi
    #     wvx=np.sqrt((self.k*self.dx)**2.+(self.l*self.dy)**2.)
    #     self.filtr = np.exp(-self.filterfac*(wvx-cphi)**4.)
    #     self.filtr[wvx<=cphi] = 1.
    #
    # def _filter(self, q):
    #     return self.filtr * q

    def set_U(self, U):
        """Set background zonal flow.

        Parameters
        ----------

        U : number
            Upper layer flow. Units meters.
        """
        self.Ubg = np.asarray(U)[np.newaxis, ...]

    def _calc_diagnostics(self):
        # here is where we calculate diagnostics
        if (self.t>=self.dt) and (self.tc%self.taveints==0):
            self._increment_diagnostics()

    ### All the diagnostic stuff follows. ###
    def _calc_cfl(self):
        return np.abs(
            np.hstack([self.u + self.Ubg, self.v])
        ).max()*self.dt/self.dx

    # calculate KE: this has units of m^2 s^{-2}
    def _calc_ke(self):
        ke = .5*self.spec_var(self.wv*self.ph)
        return ke.sum()

    # calculate eddy turn over time
    # (perhaps should change to fraction of year...)
    def _calc_eddy_time(self):
        """ estimate the eddy turn-over time in days """
        ens = .5*self.H * self.spec_var(self.wv2*self.ph)
        return 2.*pi*np.sqrt( self.H / ens ) / year

    # def _calc_derived_fields(self):
    #     self.p = self.ifft2( self.ph)
    #     self.xi =self.ifft2( -self.wv2*self.ph)
    #     self.Jptpc = -self.advect(self.p,self.u,self.v)
    #     self.Jpxi = self.advect(self.xi, self.u, self.v)
